# Copyright (c) 2016-2019 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import setuptools

setuptools.setup(setup_requires=['pbr>=5'], pbr=True)
